local LIP = nil
local libSotA_Commands = nil

local fxstop_core = {
	version = "0.03",
	scriptName = "FXStop",
	userSettings = {},
	fx_map = {}
}

fxstop_core.loadSettings = function()
	
	-- ensure the user setting file exists.  this will create it if it doesnt.
	local file = io.open(ShroudLuaPath .. '/fxstop/user.ini', "r")
	if file == nil then 
		file = io.open(ShroudLuaPath .. '/fxstop/user.ini', "w")
		if file then 
			file:write()
			file:close() 
		end
	else
		file:close()
	end
	
	fxstop_core.userSettings = LIP().load(ShroudLuaPath .. '/fxstop/user.ini')

	if (fxstop_core.userSettings["settings"] == nil) then
		fxstop_core.userSettings["settings"] = {
			debugLevel = 0
		}		
		
		LIP().save(ShroudLuaPath .. '/fxstop/user.ini', fxstop_core.userSettings)
	end
	
	if (fxstop_core.userSettings["visual_fx"] == nil) then
		fxstop_core.userSettings["visual_fx"] = {
			lightningStorm = 0,
			ringOfFire = 0,
			iceField = 0,
			soothingRain = 0,
			coneOfCold = 0,
			meteorStorm = 0,
			solarFlare = 0,
			earthquake = 0
		}		
		
		LIP().save(ShroudLuaPath .. '/fxstop/user.ini', fxstop_core.userSettings)
	end

	-- ensure the fx-map file exists.  this will create it if it doesnt.
	local file = io.open(ShroudLuaPath .. '/fxstop/fx-map.ini', "r")
	if file == nil then 
		file = io.open(ShroudLuaPath .. '/fxstop/fx-map.ini', "w")
		if file then 
			file:write()
			file:close() 
		end
	else
		file:close()
	end

	fxstop_core.fx_map = LIP().load(ShroudLuaPath .. '/fxstop/fx-map.ini')

	if (fxstop_core.fx_map["fx-map"] == nil) then
		
		fxstop_core.fx_map["fx-map"] = {}
	
		fxstop_core.fx_map["fx-map"]["VFX_LightningStorm"] = "lightningStorm"
		fxstop_core.fx_map["fx-map"]["RingOfFire_AOE"] = "ringOfFire"
		fxstop_core.fx_map["fx-map"]["IceField_Success"] = "iceField"
		fxstop_core.fx_map["fx-map"]["SoothingRain_Lasting"] = "soothingRain"
		fxstop_core.fx_map["fx-map"]["Cone_of_Cold_Success"] = "coneOfCold"
		fxstop_core.fx_map["fx-map"]["MeteorStorm"] = "meteorStorm"
		fxstop_core.fx_map["fx-map"]["SolarFlare"] = "solarFlare"
		fxstop_core.fx_map["fx-map"]["Earthquake_Success"] = "earthquake"
		fxstop_core.fx_map["fx-map"]["Unicorn_Rainbow_Attack"] = "unicornRainbowAttack"
		fxstop_core.fx_map["fx-map"]["Nightmare_Flame_Attack"] = "nightmareFlameAttack"
		fxstop_core.fx_map["fx-map"]["Fireball"] = "fireball"
		fxstop_core.fx_map["fx-map"]["Fireball_Start"] = "fireball"
		fxstop_core.fx_map["fx-map"]["Fireball_End"] = "fireball"
	
		LIP().save(ShroudLuaPath .. '/fxstop/fx-map.ini', fxstop_core.fx_map)
	end
	
	SES.invoke("fxstop.onUserSettingsLoad", fxstop_core.userSettings, fxstop_core.fx_map)
end

fxstop_core.ChangeFXStatus = function(fxName, enabled)
	
	fxstop_core.userSettings["visual_fx"][fxName] = enabled
	
	LIP().save(ShroudLuaPath .. '/fxstop/user.ini', fxstop_core.userSettings)
	SES.invoke("fxstop.onFXStatusChange", fxName, enabled, fxstop_core.userSettings)
end

fxstop_core.ChangeDebugLevel = function(debugLevel)
	fxstop_core.userSettings["settings"]["debugLevel"] = debugLevel
	
	LIP().save(ShroudLuaPath .. '/fxstop/user.ini', fxstop_core.userSettings)
	SES.invoke("fxstop.onDebugLevelChange", debugLevel, fxstop_core.userSettings)
end

fxstop_core.initializeCommands = function()
	libSotA_Commands.command.add("fxstop", function(source, action, ...)
			
		if (action == "stop") then
			fxstop_core.ChangeFXStatus(..., 1)
		elseif (action == "allow") then
			fxstop_core.ChangeFXStatus(..., 0)
		elseif (action == "list") then
			SES.invoke("fxstop.onListKeys", fxstop_core.userSettings["visual_fx"])
		elseif (action == "debug") then
			fxstop_core.ChangeDebugLevel(...)
		elseif (action == "help") then
		
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("Commands are entered into local chat in the following format:")
			ConsoleLog("[FFFF00]!fxstop <command> <parameter>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("Available Commands")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To list available FX keys and their enabled status:")
			ConsoleLog("[FFFF00]!fxstop list[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To prevent an FX:")
			ConsoleLog("[FFFF00]!fxstop stop <FX-key>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
			ConsoleLog("To allow an FX:")
			ConsoleLog("[FFFF00]!fxstop allow <FX-key>[-]")
			ConsoleLog("[808080]---------------------------------------------[-]")
		end	
	end)
	
	libSotA_Commands.command.restrict("fxstop", "Local", ShroudGetPlayerName())
	
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Welcome to version " .. fxstop_core.version .. " of the " .. fxstop_core.scriptName .. " lua addon for Shroud of the Avatar")
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("This lua addon can be found on the SotA forum at [FFFF00]https://www.shroudoftheavatar.com/forum/index.php?threads/fxstop-addon-to-prevent-certain-fx.172956/[-]")
	ConsoleLog("This lua addon utilizes the SotA Event System (SES) [FFFF00]https://www.shroudoftheavatar.com/forum/index.php?threads/sota-event-system-addon.172117/[-]")
	ConsoleLog("This lua addon utilizes a customized version of libsota.  Thank you to the original author CatweazleX.")
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("Commands use the format [FFFF00]!fxstop <command> <parameter>[-] and are entered into local chat.")
	ConsoleLog("Use [FFFF00]!fxstop help[-] in local chat to see a list of available commands.")
	ConsoleLog("[808080]---------------------------------------------[-]")
	
end

fxstop_core.onStart = function()
	LIP = SES.utility.loadmodule("fxstop/modules/LIP.lua")
	libSotA_Commands = SES.utility.loadmodule("fxstop/modules/libsotamod/libsota-commands")()

	fxstop_core.loadSettings()	
	fxstop_core.initializeCommands()
end

fxstop_core.onShroudPlayFX = function(fxName, isStart)
	
	if (isStart == false or isStart == nil) then return end
	
	if (fxstop_core.userSettings["settings"]["debugLevel"] != nil and tonumber(fxstop_core.userSettings["settings"]["debugLevel"]) > 0) then
		ConsoleLog(string.format("FX is playing (%s)", fxName))
	end
	
	local fxKeyIndex = string.gsub(fxName, "%(Clone%)", "")
	
	if (fxstop_core.fx_map["fx-map"][fxKeyIndex] == nil) then return end

	local fxKey = fxstop_core.fx_map["fx-map"][fxKeyIndex]

	if (tonumber(fxstop_core.userSettings["visual_fx"][fxKey]) == 1) then
		ShroudStopFX(fxName)
		if (fxstop_core.userSettings["settings"]["debugLevel"] != nil and tonumber(fxstop_core.userSettings["settings"]["debugLevel"]) > 0) then
			ConsoleLog(string.format("FX is stopping (%s)", fxName))
		end
	end	
end

fxstop_core.onShroudPlaySound = function(soundName, isStart)

	--ConsoleLog(string.format("Sound is playing (%s)", soundName))
	if (isStart == false or isStart == nil) then return end
--	ConsoleLog(string.format("Sound is playing (%s)", soundName))
	
end

fxstop_core.onListKeys = function(keyList)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("FXStop Status List")
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("[808080]FX Key | Block Enabled[-]")
	ConsoleLog("[808080]---------------------------------------------[-]")

	for fxKey, enabled in pairs(keyList) do
		if (enabled == 1) then 
			ConsoleLog("[FFA500]" .. fxKey .. "[-] | Block [00FF00]enabled[-]")
		else
			ConsoleLog("[FFA500]" .. fxKey .. "[-] | Block [FF0000]disabled[-]")
		end
	end	

	ConsoleLog("[808080]---------------------------------------------[-]")
end

fxstop_core.onFXStatusChange = function(fxName, enabled, settings)
	ConsoleLog("[808080]---------------------------------------------[-]")
	if (enabled) then
		ConsoleLog("Block on FX [FFFF00]" .. fxName .. "[-] has been [00FF00]enabled[-].")
	else
		ConsoleLog("Block on FX [FFFF00]" .. fxName .. "[-] has been [FF0000]disabled[-].")
	end
	ConsoleLog("[808080]---------------------------------------------[-]")
end

fxstop_core.onDebugLevelChange = function(debugLevel, settings)
	ConsoleLog("[808080]---------------------------------------------[-]")
	ConsoleLog("[FFA500]DebugLevel[-] has been changed to [00FF00]" .. debugLevel .. "[-].")
	ConsoleLog("[808080]---------------------------------------------[-]")
end

function fxstop_core:initialize()

	SES.on("ShroudOnStart", fxstop_core.onStart)
	SES.on("ShroudOnPlayFX", fxstop_core.onShroudPlayFX)
	SES.on("ShroudOnPlaySound", fxstop_core.onShroudPlaySound)
	SES.on("fxstop.onFXStatusChange", fxstop_core.onFXStatusChange)
	SES.on("fxstop.onListKeys", fxstop_core.onListKeys)
	SES.on("fxstop.onDebugLevelChange", fxstop_core.onDebugLevelChange)

	return self
end

return fxstop_core:initialize()