--[[
 * libsota-ui.lua
 * Copyright (C) 2022 Garritt Grandberg (aka Tirrag/peace-n-love)

 This file is a port of the ui functionality from libsota put into a module format.

   libsota is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   libsota is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.";
]]


-- TODO: validate for SES

local libsota_ui = {
	ui_guiObjectList = {},
	ui_drawListChanged = false	
}

libsota_ui.ui = {
	version = "0.4.8",
	guiObject = {
		add = function(left, top, width, height, drawFunc)
			local index = #libsota_ui.ui_guiObjectList + 1
			libsota_ui.ui_guiObjectList[index] = {
				rect = { left = left or 0, top = top or 0, width = width or 0, height = height or 0 },
				visible = false,
				shownInScene = true,
				shownInLoadScreen = false,
				zIndex = 0,
			}
			local obj = {
				guiDrawFunc = drawFunc,
				guiObjectId = index,
				rect = libsota_ui.ui_guiObjectList[index].rect,
			}
			setmetatable(obj, {__index = libsota_ui.ui.guiObject})
			setmetatable(libsota_ui.ui_guiObjectList[index], {__index = obj})
			return obj
		end,
		remove = function(obj)
			libsota_ui.ui_guiObjectList[obj.guiObjectId] = nil
			obj = nil
			libsota_ui.ui_drawListChanged = true
		end,
		shownInScene = function(obj, shown)
			if shown ~= nil and shown ~= libsota_ui.ui_guiObjectList[obj.guiObjectId].shownInScene then
				libsota_ui.ui_guiObjectList[obj.guiObjectId].shownInScene = shown
				libsota_ui.ui_drawListChanged = true
			end
			return libsota_ui.ui_guiObjectList[obj.guiObjectId].shownInScene
		end,
		shownInLoadScreen = function(obj, shown)
			if shown ~= nil and shown ~= libsota_ui.ui_guiObjectList[obj.guiObjectId].shownInLoadScreen then
				libsota_ui.ui_guiObjectList[obj.guiObjectId].shownInLoadScreen = shown
				libsota_ui.ui_drawListChanged = true
			end
			return libsota_ui.ui_guiObjectList[obj.guiObjectId].shownInLoadScreen
		end,
		zIndex = function(obj, zIndex)
			if zIndex ~= nil and zIndex ~= libsota_ui.ui_guiObjectList[obj.guiObjectId].zIndex then
				libsota_ui.ui_guiObjectList[obj.guiObjectId].zIndex = zIndex
				libsota_ui.ui_drawListChanged = true
			end
			return libsota_ui.ui_guiObjectList[obj.guiObjectId].zIndex
		end,
		visible = function(obj, visible)
			if visible ~= nil and visible ~= libsota_ui.ui_guiObjectList[obj.guiObjectId].visible then
				libsota_ui.ui_guiObjectList[obj.guiObjectId].visible = visible
				libsota_ui.ui_drawListChanged = true
			end
			return libsota_ui.ui_guiObjectList[obj.guiObjectId].visible
		end,
		toggle = function(obj)
			obj:visible(not obj:visible())
		end,
		rect = function(obj, rect)
			if type(rect) == "table" then
				if rect.left then obj.rect.left = rect.left end
				if rect.top then obj.rect.top = rect.top end
				if rect.width then obj.rect.width = rect.width end
				if rect.height then obj.rect.height = rect.height end
			end
			return obj.rect
		end,
		moveTo = function(obj, x, y)
			obj.rect.left = x
			obj.rect.top = y
		end,
		moveBy = function(obj, x, y)
			obj.rect.left = obj.rect.left + x
			obj.rect.top = obj.rect.top + y
		end,
		resizeTo = function(obj, w, h)
			obj.rect.width = w
			obj.rect.height = h
		end,
		resizeBy = function(obj, x, y)
			obj.rect.width = obj.rect.width + x
			obj.rect.height = obj.rect.height + y
		end,
	},
	
	
	label = {
		add = function(left, top, width, height, caption)
			local l = libsota_ui.ui.guiObject.add(left, top, width, height, libsota_ui.ui.label.draw)
			l.caption = caption or ""
			
			setmetatable(l, {__index = libsota_ui.ui.label})
			return l
		end,
		draw = function(l)
			ShroudGUILabel(l.rect.left, l.rect.top, l.rect.width, l.rect.height, l.caption)
		end,
	},
	
	texture = {
		_loaded = {},
		add = function(left, top, filename, clamped, scaleMode, width, height)
			
			if not libsota_ui.ui.texture._loaded[tostring(filename)] then
				local tid = ShroudLoadTexture(filename, clamped or false)
				if tid < 0 then
					print("libsota_ui.ui.texture: Unable to load texture with filename '"..filename.."'")
					return nil
				end
				local tw, th = ShroudGetTextureSize(tid)
				libsota_ui.ui.texture._loaded[tostring(filename)] = {
					filename = filename,
					id = tid,
					width = tw,
					height = th,
					clamped = clamped or false,
				}
			end
			local tex = libsota_ui.ui.texture._loaded[tostring(filename)]

			local t = libsota_ui.ui.guiObject.add(left, top, width or tex.width, height or tex.height, libsota_ui.ui.texture.draw)
			t.texture = tex
			t.scaleMode = scaleMode or -1

			setmetatable(t, {__index = libsota_ui.ui.texture})
			return t
		end,
		draw = function(t)
			ShroudDrawTexture(t.rect.left, t.rect.top, t.rect.width, t.rect.height, t.texture.id, t.scaleMode)
		end,
		clone = function(texture)
			local t = libsota_ui.ui.guiObject.add(texture.rect.left, texture.rect.top, texture.rect.width, texture.rect.height, texture.guiDrawFunc)
			t.texture = texture.texture
			t.scaleMode = texture.scaleMode
			t.shownInScene = texture.shownInScene
			t.shownInLoadScreen = texture.shownInLoadScreen
			t.zIndex = texture.zIndex
						
			setmetatable(t, {__index = libsota_ui.ui.texture})
			return t
		end,
		clamp = function(texture, clamped)
			if clamped ~= nil and clamped ~= texture.texture.clamped then
				texture.texture.clamped = clamped
				ShroudSetTextureClamp(texture.texture.id, texture.texture.clamped)
			end
			return texture.texture.clamped
		end,
		scaleMode = function(texture, scaleMode)
			if scaleMode ~= nil then
				texture.scaleMode = scaleMode
			end
			return texture.scaleMode
		end,
	},
	
	guiButton = {
		add = function(left, top, texture, caption, tooltip, width, height)
			local b = libsota_ui.ui.guiObject.add(left, top, width, height, libsota_ui.ui.guiButton.draw)
			b.caption = caption or ""
			b.tooltip = tooltip or ""
			-- b.mode = false -- mode = repeat, click, off/false

			if type(texture) == "table" and texture.texture and texture.texture.id then
				b.texture = texture
			else
				print("libsota_ui.ui.guiButton: "..type(texture).." is not a texture object.")
				return nil
			end
			
			setmetatable(b, {__index = libsota_ui.ui.guiButton})
			return b
		end,
		draw = function(b)
			ShroudButton(b.rect.left, b.rect.top, b.rect.width, b.rect.height, b.texture.texture.id, b.caption, b.tooltip)
		end,
		onClick = function(button, callback)
			button.callback = callback
			button.guiDrawFunc = function(b)
				if ShroudButton(b.rect.left, b.rect.top, b.rect.width, b.rect.height, b.texture.texture.id, b.caption, b.tooltip) then
					b.callback(ShroudMouseX - b.rect.left, ShroudMouseY - b.rect.top)
				end
			end
		end,
		onRepeat = function(button, callback)
			button.callback = callback
			button.guiDrawFunc = function(b)
				if ShroudButtonRepeat(b.rect.left, b.rect.top, b.rect.width, b.rect.height, b.texture.texture.id, b.caption, b.tooltip) then
					b.callback(ShroudMouseX - b.rect.left, ShroudMouseY - b.rect.top)
				end
			end
		end,
	},	
}

setmetatable(libsota_ui.ui.label, {__index = libsota_ui.ui.guiObject})
setmetatable(libsota_ui.ui.texture, {__index = libsota_ui.ui.guiObject})
setmetatable(libsota_ui.ui.guiButton, {__index = libsota_ui.ui.guiObject})

local ui_drawInScene = {}
local ui_drawInLoadScreen = {}
libsota_ui.ui_client_ts = os.time()

local function ui_buildDrawList()
	local list = libsota_ui.ui_guiObjectList
	local ds = {}
	local dl = {}

	for _,o in next, list do
		if o.visible and o.shownInScene then
			ds[#ds + 1] = o
		elseif o.visible and o.shownInLoadScreen then
			dl[#dl + 1] = o
		end
	end

	table.sort(ds, function(a, b) return a.zIndex < b.zIndex end)
	table.sort(dl, function(a, b) return a.zIndex < b.zIndex end)
	
	ui_drawInScene = ds
	ui_drawInLoadScreen = dl
end

libsota_ui.onUpdate = function()

	local ts = os.time()
	
	if libsota_ui.ui_drawListChanged then
		libsota_ui.ui_drawListChanged = false
		ui_buildDrawList()
	end

	libsota_ui.ui_client_ts = ts
end

libsota_ui.onGUI = function()
	local ts = os.time()
	local isHitching = ts - libsota_ui.ui_client_ts >= 2
	local isLoading = ts - libsota_ui.ui_client_ts >= 5

	if not isHitching then
		local list = ui_drawInScene
		if isLoading then list = ui_drawInLoadScreen end
		
		for _,o in next, list do
			o:guiDrawFunc()
			if os.time() - ts > 0.01 then break end
		end
	end
end

libsota_ui.onStart = function()
end

function libsota_ui.ui:initialize()
	SES.on("ShroudOnStart", libsota_ui.onStart)
	SES.on("ShroudOnUpdate", libsota_ui.onUpdate)
	SES.on("ShroudOnGUI", libsota_ui.onGUI)
	
	return self
end

return libsota_ui.ui:initialize()