--[[
 * libsota.lua
 * Copyright (C) 2022 Garritt Grandberg (aka Tirrag/peace-n-love)

 This file is a port of the core libsota functionality from libsota put into a module format.

   libsota is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   libsota is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.";
]]


-- TODO: validate for SES

local libsota = {
	version = "0.4.8",
}

libsota.client = {
	timeStarted = os.time(),
	timeToLoad = 0,
	timeInGame = 0,
	timeDelta = 0,
	fps = 0,
	accuracy = 0,
	isHitching = false,
	isLoading = false,
	_statEnum = {},
	_statDescr = {},
	screen = {
		width = 0,
		height = 0,
		isFullScreen = false,
		aspectRatio = 0,
		_fsmul = 1,
	},
	api = {
		luaVersion = "",
		luaPath = "",
		list = {},
		isImplemented = function(name)
			return libsota.client.api.list[name] ~= nil
		end
	},
	mouse = {
		button = {},
		x = 0,
		y = 0,
	},
	window = {
		paperdoll = { name = "paperdoll", open = false, left = 0, top = 0, width = 267, height = 487 }
	}
}

libsota.scene = {
	name = "none",
	maxPlayer = 0,
	isPvp = false,
	isPot = false,
	timeInScene = 0,
	timeToLoad = 0,
	timeStarted = 0,
}

libsota.player = {
	caption = "",
	name = "none",
	flag = "",
	isPvp = false,
	isAfk = false,
	isGod = false,
	isMoving = false,
	isStill = false,
	lastMoved = os.time(),
	location = { x = 0, y = 0 , z = 0, scene = scene },
	health = { current = 0, max = 0, percentage = 0 },
	focus = { current = 0, max = 0, percentage = 0 },
	xp = { producer = 0, adventurer = 0 },
	inventory = {},
	stat = function(index)
	
		local ret = {
			number = -1,
			name = "invalid",
			value = -999,
			description = "",
		}
		if tonumber(index) == nil then
			index = libsota.client._statEnum[index]
		else
			index = tonumber(index)
		end

		if index and index <= #libsota.client._statEnum then
			ret.number = index
			ret.name = libsota.client._statEnum[index]
			ret.value = ShroudGetStatValueByNumber(index)
			ret.description = libsota.client._statDescr[index]
		end
		return ret
	end,
}

-- libsota events
libsota.events = {
	onInit = function(callback) return SES.on("libsota._init", callback) end,
	onStart = function(callback) return SES.on("libsota._start", callback) end,
	onUpdate = function(callback) 
		return SES.on("libsota._update", callback) 
	end,
	onConsoleInput = function(callback) return SES.on("libsota._consoleInput", callback) end,
	onConsoleCommand = function(callback) return SES.on("libsota._consoleCommand", callback) end,
	onSceneChanged = function(callback) return SES.on("libsota._sceneChanged", callback) end,
	onPlayerChanged = function(callback) return SES.on("libsota._playerChanged", callback) end,
	onPlayerMove = function(callback) return SES.on("libsota._playerMove", callback) end,
	onPlayerMoveStart = function(callback) return SES.on("libsota._playerMoveStart", callback) end, -- depricated
	onPlayerMoveStop = function(callback) return SES.on("libsota._playerMoveStop", callback) end, -- depricated
	onPlayerIsStill = function(callback) return SES.on("libsota._playerIsStill", callback) end,
	onPlayerDamage = function(callback) return SES.on("libsota._playerDamage", callback) end,
	onPlayerInventory = function(callback) return SES.on("libsota._playerInventory", callback) end,
	onClientWindow = function(callback) return SES.on("libsota._clientWindow", callback) end,
	onClientIsHitching = function(callback) return SES.on("libsota._clientIsHitching", callback) end,
	onClientIsLoading = function(callback) return SES.on("libsota._clientIsLoading", callback) end,
	onMouseMove = function(callback) return SES.on("libsota._mouseMove", callback) end,
	onMouseButton = function(callback) return SES.on("libsota._mouseButton", callback) end,
}


libsota.utility = {
	
	module = {
		list = {},
		add = function(modulename, prefix)
			local part = {}
			for p in modulename:gmatch("[^\\/.]+") do part[#part + 1] = p end
			if part[#part] == "lua" then part[#part] = nil end
			modulename = table.concat(part, ".")
			if not libsota.core.module.list[modulename] then
				local filename = table.concat(part, "/")..".lua"
				local f = nil
				for _,inc in next, {"/","/lib/","/share/","/bin/","/include/"} do
					f = io.open(ShroudLuaPath..inc..filename)
					if f then break end
				end
				if not f then
					print("libsota.core.module: '",filename,"' not found")
					return nil
				end
				local d = f:read("*a")
				f:close()
				local m,e = loadsafe(d, modulename, "bt", _G)
				if not m then
					print("libsota.core.module: '",modulename,"': "..e)
					return nil
				end
				libsota.core.module.list[modulename] = {}
				libsota.core.module.list[modulename][0] = m
			end
			if prefix then
				if prefix == true or prefix == "" then
					prefix = modulename
				else
					part = {}
					for p in prefix:gmatch("[^\\/.]+") do part[#part + 1] = p end
					if part[#part] == "lua" then part[#part] = nil end
					prefix = table.concat(part, ".")
				end
				if #part == 1 then
					_G[part[1]] = libsota.core.module.list[modulename][0]
				else
					local tree = {}
					local leave = tree
					for i=2,#part-1 do
						if not leave[part[i]] then leave[part[i]] = {} end
						leave = leave[part[i]]
					end
					leave[part[#part]] = libsota.core.module.list[modulename][0]
					_G[part[1]] = tree
				end
				libsota.core.module.list[modulename][#libsota.core.module.list[modulename] + 1] = part
			end
			return libsota.core.module.list[modulename][0]
		end,
		get = function(modulename)
			local part = {}
			for p in modulename:gmatch("[^\\/.]+") do part[#part + 1] = p end
			if part[#part] == "lua" then part[#part] = nil end
			modulename = table.concat(part, ".")
			return libsota.core.module.list[tostring(modulename)][0]
		end,
		remove = function(modulename)
			local part = {}
			for p in modulename:gmatch("[^\\/.]+") do part[#part + 1] = p end
			if part[#part] == "lua" then part[#part] = nil end
			modulename = table.concat(part, ".")
			libsota.core.module.list[tostring(modulename)][0] = nil
			for _,prefix in next, libsota.core.module.list[tostring(modulename)] do
				if #prefix == 1 then
					_G[prefix[1]] = nil
				else
					local tree = _G[prefix[1]]
					local leave = tree
					for i=2,#prefix-1 do
						if leave[prefix[i]] then leave = leave[prefix[i]] end
					end
					leave[prefix[#prefix]] = nil
					_G[prefix[1]] = tree
				end
			end
			libsota.core.module.list[tostring(modulename)] = nil
		end,
		invoke = function(modulename, callback, ...)
			local part = {}
			for p in modulename:gmatch("[^\\/.]+") do part[#part + 1] = p end
			if part[#part] == "lua" then part[#part] = nil end
			modulename = table.concat(part, ".")
			if libsota.core.module.list[modulename] then
				if callback then
					return libsota.core.module.list[modulename][0][tostring(callback)](...)
				else
					return libsota.core.module.list[modulename][0](...)
				end
			else
				print("libsota.core.module: ",modulename," not found")
			end
		end,
	},
	
	verbosity = 0,
	consoleLog = function(message, verbosity)
		if not verbosity then verbosity = 0 end
		if libsota.utility.verbosity >= verbosity then
			for l in message:gmatch("[^\n]+") do
				ShroudConsoleLog(l)
			end
		end
	end,
}


-- internal
libsota.ui_initialized = false
libsota.ui_client_ts = os.time()
libsota.ui_client_frame = 0
local ui_callback_queue = {}

local ui_inventory_quantity = {}
local ui_inventory_durability = {}

local function ui_getPlayerName()
	if libsota.player.caption ~= ShroudGetPlayerName() then
		libsota.player.caption = ShroudGetPlayerName()
		libsota.player.name = libsota.player.caption
		libsota.player.flag = ""
		libsota.player.isPvp = false
		if libsota.player.name:byte(#libsota.player.name) == 93 then
			libsota.player.name, libsota.player.flag = libsota.player.name:match("^(.-)(%[.-%])$")
			libsota.player.isPvp = libsota.player.flag and libsota.player.flag:find("PVP") ~= nil
			libsota.player.isAfk = libsota.player.flag and libsota.player.flag:find("AFK") ~= nil
			libsota.player.isGod = libsota.player.flag and libsota.player.flag:find("God") ~= nil
		end
		SES.invoke("libsota._playerChanged", "caption")
	end
end

local function ui_getPlayerChange()
	local loc = { x = ShroudPlayerX, y = ShroudPlayerY, z = ShroudPlayerZ, scene = scene }
	local isMoving = loc.x ~= libsota.player.location.x or loc.y ~= libsota.player.location.y or loc.z ~= libsota.player.location.z
	if isMoving then libsota.player.lastMoved = os.time() end
	local isStill = os.time() - libsota.player.lastMoved > 5
	local invoke = isStill ~= libsota.player.isStill or isMoving ~= libsota.player.isMoving
	libsota.player.isMoving = isMoving
	libsota.player.isStill = isStill
	libsota.player.location = loc
	if invoke then
		if isMoving then
			SES.invoke("libsota._playerMoveStart") -- depricated
			SES.invoke("libsota._playerMove", "start", loc)
			SES.invoke("libsota._playerChanged", "moving")
		elseif isStill then
			SES.invoke("libsota._playerIsStill") -- depricated?
			SES.invoke("libsota._playerMove", "stillness", loc) -- hm?
			SES.invoke("libsota._playerChanged", "stillness")
		else
			SES.invoke("libsota._playerMoveStop") -- depricated
			SES.invoke("libsota._playerMove", "stop", loc)
			SES.invoke("libsota._playerChanged", "standing")
		end
	end
	if isMoving then
		SES.invoke("libsota._playerMove", "move", loc)
		SES.invoke("libsota._playerChanged", "location", loc)
	end
	local health = { max = ShroudGetStatValueByNumber(30), current = ShroudPlayerCurrentHealth, percentage = 0 }
	local focus =  { max = ShroudGetStatValueByNumber(27), current = ShroudPlayerCurrentFocus, percentage = 0 }
	health.percentage = health.current / health.max
	focus.percentage = focus.current / focus.max
	invoke = libsota.player.health.current ~= health.current or libsota.player.focus.current ~= focus.current
	if invoke then
		SES.invoke("libsota._playerDamage", health, focus)
		SES.invoke("libsota._playerChanged", "damage", health, focus)
	end
	libsota.player.health = health
	libsota.player.focus = focus
end

local function ui_getPlayerInventory()
	local inventory = ShroudGetInventory()
	local ret1 = {}
	local ret2 = {}
	local ret3 = {}
	for _, item in next, inventory do
		item = table.pack(item)
		
		ret1[#ret1 + 1] = {
			name = item[1],
			durability = item[2],
			primaryDurability = item[3],
			maxDurability = item[4],
			weight = item[5],
			quantity = item[6],
			createdBy = item[7],
			creationTime = item[8],
			creationScene = item[9],
			lengthOrSize = item[10],
			value = item[11],
		}
		if not ret2[tostring(item[1])] then
			ret2[tostring(item[1])] = 0
		end
		ret2[tostring(item[1])] = ret2[tostring(item[1])] + item[6]
	end
	local cs = { n = 0 }
	for n, q in next, ui_inventory_quantity do
		if not ret2[n] then
			--print("Gegenstand entfernt: "..n.."Menge: "..q)
			cs[n] = { quantity = q, action = "removed" }
			cs.n = cs.n + 1
		elseif ret2[n] ~= q then
			--print("Gegenstand geändert: "..n.."Menge: "..(ret2[n] - q))
			cs[n] = { quantity = ret2[n] - q, action = "changed" }
			cs.n = cs.n + 1
		end
	end
	for n, q in next, ret2 do
		if not ui_inventory_quantity[n] then
			--print("Gegenstand zugefügt: "..n.."Menge: "..q)
			cs[n] = { quantity = q, action = "added" }
			cs.n = cs.n + 1
		end
	end
	
	libsota.player.inventory = ret1
	ui_inventory_quantity = ret2

	if cs.n > 0 then
		cs.n = nil
		SES.invoke("libsota._playerInventory", cs)
		SES.invoke("libsota._playerChanged", "inventory", cs)
	end
end

local function ui_getSceneName()
	if libsota.scene.name ~= ShroudGetCurrentSceneName() then
		libsota.scene.name = ShroudGetCurrentSceneName()
		libsota.scene.maxPlayer = ShroudGetCurrentSceneMaxPlayerCount()
		libsota.scene.isPvp = ShroudGetCurrentSceneIsPVP()
		libsota.scene.isPot = ShroudGetCurrentSceneIsPOT()
		libsota.scene.timeStarted = os.time()
		SES.invoke("libsota._sceneChanged")
	end
end

local function ui_initialize()

	libsota.client.api.luaPath = ShroudLuaPath
	libsota.client.api.luaVersion = _G._MOONSHARP.luacompat
	for i,v in next, _G do
		if i:find("^Shroud") then libsota.client.api.list[tostring(i)] = type(v) end
	end

	libsota.client.timeToLoad = libsota.ui_client_ts - libsota.client.timeStarted
	libsota.client.screen.width = ShroudGetScreenX()
	libsota.client.screen.height = ShroudGetScreenY()
	libsota.client.screen.isFullScreen = ShroudGetFullScreen()
	libsota.client.screen.aspectRatio = libsota.client.screen.width / libsota.client.screen.height
	libsota.client.screen._fsmul = libsota.client.screen.height / 1080
	for i=0, ShroudGetStatCount()-1, 1 do
		local name = ShroudGetStatNameByNumber(i)
		libsota.client._statEnum[tostring(name)] = i
		libsota.client._statEnum[i] = name
		libsota.client._statDescr[i] = ShroudGetStatDescriptionByNumber(i)
	end
		
	ui_getSceneName()
	ui_getPlayerName()
	ui_getPlayerChange()
	ui_getPlayerInventory()
end

local function ui_queue(callback, ...)
	local index = #ui_callback_queue + 1

	if type(callback) == "function" then
		ui_callback_queue[index] = {
			callback = callback,
			userdata = table.pack(...),
		}	
	end
end

-- shroud timers
local ui_timer_ts_fast = os.time()
local ui_timer_ts_slow = os.time()
local ui_timer_internal_last_ts = nil

local function ui_timer_internal()
	local ts = os.time()

	-- see if we need to run yet
	if (ui_timer_user_last_ts != nil and (ts - ui_timer_user_last_ts) < 0.01) then return end

	if ts - ui_timer_ts_fast > 0.240 then
		ui_timer_ts_fast = ts

		ui_getPlayerChange()
	end
	
	if ts - ui_timer_ts_slow > 2 then
		ui_timer_ts_slow = ts
		
		-- watch player xp
		libsota.player.xp.producer = ShroudGetPooledProducerExperience()
		libsota.player.xp.adventurer = ShroudGetPooledAdventurerExperience()

		-- watch player inventory
		ui_getPlayerInventory()
	end


	-- queued callbacks

	for i,q in next, ui_callback_queue do
		q.callback(unpack(q.userdata))
		ui_callback_queue[i] = nil
	end
	
	
	-- watch character sheet window (realtime)

	local wo = ShroudIsCharacterSheetActive()
	if wo then
		local wx, wy = ShroudGetCharacterSheetPosition(); wx = wx - 860 -- bug
		if wx ~= libsota.client.window.paperdoll.left or wy ~= libsota.client.window.paperdoll.top then
			libsota.client.window.paperdoll.left, libsota.client.window.paperdoll.top = wx, wy
			SES.invoke("libsota._clientWindow", "moved", libsota.client.window.paperdoll)		
		end
	end
	if wo ~= libsota.client.window.paperdoll.open then
		libsota.client.window.paperdoll.open = wo
		if wo then
			SES.invoke("libsota._clientWindow", "opened", libsota.client.window.paperdoll)
		else
			SES.invoke("libsota._clientWindow", "closed", libsota.client.window.paperdoll)
		end
	end
	
	ui_timer_internal_last_ts = ts
end

local button_ts = {} -- howto: clean up
local button_ts2 = {}

-- Shroud events
-- setup event callbacks to ShroudOn functions

function libsota:initialize()
	SES.on("ShroudOnStart", libsota.onStart)
	SES.on("ShroudOnUpdate", libsota.onUpdate)
	SES.on("ShroudOnGUI", libsota.onGUI)
	SES.on("ShroudOnConsoleInput", libsota.onConsoleInput)
	
	return self
end


libsota.onStart = function()
	--ShroudUseLuaConsoleForPrint(true)
	--ShroudConsoleLog(_G._MOONSHARP.banner)
	--ShroudConsoleLog("LUA Version: ".._G._MOONSHARP.luacompat)
end

libsota.onUpdate = function()

	local ts = os.time()

	-- init
	if not libsota.ui_initialized then
		if not ShroudServerTime then return end -- shroud Api not ready, yet
		--ShroudConsoleLog("Shroud Api Ready. ServerTime: "..ShroudServerTime)

		ui_initialize()
		libsota.ui_initialized = true

		SES.invoke("libsota._init")
		SES.invoke("libsota._start")
	end

	ui_timer_internal()

	-- client stats
	libsota.ui_client_frame = libsota.ui_client_frame + 1;
	if ts - libsota.ui_client_ts >= 1 then
		if libsota.client.accuracy > 10 then
			ui_getSceneName()
			ui_getPlayerName()
			libsota.client.isLoading = false
			libsota.scene.timeToLoad = libsota.client.accuracy
		end
		libsota.client.timeInGame = ShroudTime
		libsota.client.timeDelta =  ShroudDeltaTime
		libsota.client.fps = libsota.ui_client_frame
		libsota.ui_client_frame = 0
		libsota.client.accuracy = ts - libsota.ui_client_ts - 1
		libsota.ui_client_ts = ts
		libsota.client.isHitching = false
		libsota.scene.timeInScene = ts - libsota.scene.timeStarted
	end

	-- check mouse
	for i,mb in next, { "Mouse0", "Mouse1", "Mouse2", "Mouse3", "Mouse4", "Mouse5", "Mouse6" } do
		libsota.client.mouse.button[i] = nil
		if ShroudGetOnKeyDown(mb) then
			button_ts[i] = os.time()
			libsota.client.mouse.button[i] = "down"
		elseif ShroudGetKeyDown(mb) then
			if button_ts[i] != nil and os.time() - button_ts[i] >= 0.25 then
				libsota.client.mouse.button[i] = "held"
			end
		elseif ShroudGetOnKeyUp(mb) then
			libsota.client.mouse.button[i] = "up"
			if button_ts[i] != nil and os.time() - button_ts[i] <= 0.25 then
				if button_ts2[i] and os.time() - button_ts2[i] <= 0.25 then
					SES.invoke("libsota._mouseButton", "dblclicked", i, ShroudMouseX, ShroudMouseY)
					button_ts2[i] = nil
				else
					SES.invoke("libsota._mouseButton", "clicked", i, ShroudMouseX, ShroudMouseY)
					button_ts2[i] = os.time()
				end
			end
		end
	end
	if libsota.client.mouse.x ~= ShroudMouseX or libsota.client.mouse.y ~= ShroudMouseY then
		libsota.client.mouse.x, libsota.client.mouse.y = ShroudMouseX, ShroudMouseY
		SES.invoke("libsota._mouseMove", libsota.client.mouse.button, libsota.client.mouse.x, libsota.client.mouse.y)
	end
	for i,mb in next, libsota.client.mouse.button do
		SES.invoke("libsota._mouseButton", mb, i, libsota.client.mouse.x, libsota.client.mouse.y)
	end


	SES.invoke("libsota._update")	
	
end

libsota.onGUI = function()
	local ts = os.time()
	local isHitching = ts - libsota.ui_client_ts >= 2
	local isLoading = ts - libsota.ui_client_ts >= 5

	if isHitching and libsota.client.isHitching ~= isHitching then
		libsota.client.isHitching = isHitching
		SES.invoke("libsota._clientIsHitching", isHitching)
	end
	if isLoading and libsota.client.isLoading ~= isLoading then
		libsota.client.isLoading = isLoading
		SES.invoke("libsota._clientIsLoading", isLoading)
	end		
end

libsota.onConsoleInput = function(channel, sender, message)

	ui_queue(function(...)

		-- handle player flag changes
		if (channel == "Story" or channel == "System") and message:find("PvP") then ui_getPlayerName() end

		-- parse message
		local src, dst, msg = message:match("^(.-) to (.-) %[.-:%s*(.*)$")
		if sender == "" then sender = src end
		if sender == "" then sender = libsota.player.name end
		if msg:byte() == 92 or msg:byte() == 95 or msg:byte() == 33 then
			local cmd, tail = msg:match("^[\\_!](%w+)%s*(.*)$")
			local source = {
				channel = channel,
				sender = sender,
				receiver = dst,
			}
			SES.invoke("libsota._consoleCommand", source, cmd, tail)
		else
			SES.invoke("libsota._consoleInput", channel, sender, dst, msg)
		end

	end, channel, sender, message)
end

return libsota:initialize()